--[[
	This mod relocates the stars of the given galaxy map in order to form a spiral
	shape.

	The galaxy shape:
	The center of this new shape is a cluster made of rings, with a single
	system in the middle (which is Orion, if it exists).
	Attached to this cluster is a small, discontinuous transition ring (inner arms)
	which leads to the spiral arms (outer arms).
	There is exactly one spiral arm per player. These arms start out a bit thicker near
	the center and narrow down quickly as they grow outwards. The home world of the
	coresponding player is located at the tip of the spiral arm.
	By default, the length of the arm depends on how many stars are left and wether
	one arm grows too close to the edge. Regardless of this, all arms will always
	be of the exact same length and shape for all players in the game in terms of
	how the stars are arranged. The actual stars and planets are not mirrored though.
	By default, the distance between two neighbouring stars is kept at 3 parsecs or
	less, so that standard fuel cells are sufficient to expand. 

	Properties of the algorithm:
	The algorithm is designed to work with any number of players, galaxy sizes
	and number of stars, though of course there are certain limits.
	Note that the galaxy is randomly rotated which means even with the same
	settings the arm length can be different for any two games as they will
	stop growing outwards as soon as they reach the edge of the galaxy.
	By default, any blackholes, wormholes and empty systems are removed to prevent
	players from gaining early access to the center or being unable to advance
	early on.

	Changing the algorithm parameters:
	Change the CAPITAL_CASE constants below. Note that some of these can have
	dramatic impact on the overall generation. I will update this section sometime
	in a future version of the mod. In the meantime, feel free to contact me directly
	in discord or by email.

	Known issues:
	- With default parameters, small galaxies only support up to 6 players.
	  This is because there are not enough stars in a small galaxy to properly run
	  with 7 oder 9 spiral arms. An appropriate error message is shown.
	- Sometimes system names overlap in the UI. This could be compensated by increasing
	  the distance between systems though exceeding 3 parsec seems undesiareble.

	Possible TODOs:
	- Explicitly prevent formation of empty systems in the first place so that they
	  do not have to be deleted from the galaxy.
	- More code-cleanup, specifically extracting more methods from the main one.

	Last update of this mod: 20th March 2021; Tested with version 1.50.13 improved

	Questions etc. to: romanhable@web.de (Epirasque)
]]

set_seed()

DELETE_ALL_BLACKHOLES = true
DELETE_ALL_EMPTY_SYSTEMS = true
REMOVE_ALL_WORMHOLES = true

CLUSTER_ROTATION_ANGLE = random(360)
ARM_ROTATION_ANGLE = random(360)

ARM_STAR_DISTANCE_IN_COORDINATES = 90 - 2 --90 equals 3 parsecs (keeping it a bit smaller in case of rounding issues in both coordinates); 3 parsec equals standard fuel cell range
ARM_DISTANCE_TO_CLUSTER_IN_COORDINATES = 60 --60 equals 2 parsec, note that the distance to the actual systems in the inner cluster will be a bit bigger in the end!
ARM_DISTANCE_INNER_TO_FIRST_OUTER_IN_COORDINATES = 60 --distance used for the transition between inner cluster and spiral arms
ARM_DISTANCE_BETWEEN_PARTIAL_RING_STARS_IN_COORDINATES = 90 - 2 --shortest distance between two systems of the rings; can effectively be a bit bigger

ARM_INNER_CIRCUMFENCE_COVERED_FACTOR_PER_PLAYER = 35 -- 100 means the player will cover the full share of its partial ring so in total the entire ring will be covered
ARM_OUTER_LINE_COVERED_STARTING_FACTOR_PER_PLAYER = 35 -- as a rule of thumb 100 is slightly less than ARM_INNER_CIRCUMFENCE_COVERED_FACTOR_PER_PLAYER; things are different because it is not bent
ARM_OUTER_LINE_COVERED_INITIAL_ROOTED_FACTOR_LOSS_PER_PLAYER = 2 --will be reduced by 1 per iteration; its square is cumulatively subtracted from the starting factor
NR_MAX_OUTER_ARM_ITERATIONS = 100 --set to at least 1, set to a high number to effectively disable it. Aside from center cluster and a single transition layer (inner arm), this is the maximum number of iterations each outer arm may grow.

--set positive values for clockwise direction; two quadractic functions are used to model the arm growth: one from initial to intermediate and one from intermediate to final
INITIAL_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION = -12
INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION = -27
NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE = 3 --note that the first arm iteration is not directly attached to inner cluster, there is one transition iteration
FINAL_SPIRAL_ARM_CHANGE_PER_ITERATION = 0
NR_ITERATIONS_TO_REACH_FINAL_SPIRAL_ARM_ANGLE_CHANGE = 10 --should be bigger than NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE

CLUSTER_DISTANCE_BETWEEN_RING_STARS_IN_COORDINATES = 90 - 2 --90 equals 3 parsecs; this is along a circle, not a direct distance, meaning actual direct distance is a bit shorter
CLUSTER_DISTANCE_BETWEEN_RINGS_IN_COORDINATES = 75 --75 equals 2.5 parsec; shortest distance between two systems of the rings can effectively be a bit bigger
CLUSTER_SIZE_LIMIT_IN_VERTICAL_PERCENTAGE = 18 --size of the inner cluster as percentage of galaxy-height

MINIMUM_DISTANCE_FROM_GALAXY_EDGE_IN_COORDINATES = 45 --30 equals 1 parsec; arms will stop growing when any of the arms would violate this threshold

BLACK_HOLE_COLOUR = 6 --color id to identify black holes; used if black holes are deleted

function factor_a_of_negative_quadratic_function_times_thousand(y_axis_intercept, vertex_x, vertex_y)
	return 1000000 * (y_axis_intercept - vertex_y) / (vertex_x * vertex_x)
end

function arm_iteration_to_single_angle_change(arm_iteration, INITIAL_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE, FINAL_SPIRAL_ARM_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_FINAL_SPIRAL_ARM_ANGLE_CHANGE)
	local x = arm_iteration - 1
	local angle_change = 0
	if arm_iteration <= NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE then
		local y_axis_intercept = INITIAL_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION
		local vertex_x = NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE
		local vertex_y = INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION
		local factor_a = factor_a_of_negative_quadratic_function_times_thousand(y_axis_intercept, vertex_x, vertex_y)
		angle_change = factor_a * (x - vertex_x) * (x - vertex_x) / 1000000 + vertex_y
	else
		x = x - NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE
		local y_axis_intercept = INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION
		local vertex_x = NR_ITERATIONS_TO_REACH_FINAL_SPIRAL_ARM_ANGLE_CHANGE
		local vertex_y = FINAL_SPIRAL_ARM_CHANGE_PER_ITERATION
		local factor_a = factor_a_of_negative_quadratic_function_times_thousand(y_axis_intercept, vertex_x, vertex_y)
		angle_change = factor_a * (x - vertex_x) * (x - vertex_x) / 1000000 + vertex_y
	end
	return angle_change
end

--todo: cache results of arm_iteration_to_single_angle_change (or integrate)
function arm_iteration_to_angle_change(arm_iteration, INITIAL_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE, FINAL_SPIRAL_ARM_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_FINAL_SPIRAL_ARM_ANGLE_CHANGE)
	return arm_iteration_to_single_angle_change(arm_iteration, INITIAL_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE, FINAL_SPIRAL_ARM_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_FINAL_SPIRAL_ARM_ANGLE_CHANGE)
end


function parsec2distance(parsec)
	return parsec*30-2
end

function distance2parsec(distance)
	local parsec = distance/30
	--math.ceil is not available
	if distance ~= parsec2distance(parsec) then
		parsec = parsec + 1
	end
	return parsec
end

function offset_to_coordinates(origin_x, origin_y, angle, distance_from_origin_in_coordinates, sinus, cosinus)
	angle = angle % 360
	new_x = origin_x + distance_from_origin_in_coordinates * cosinus[angle] / 1000000
	new_y = origin_y + distance_from_origin_in_coordinates * sinus[angle] / 1000000
	return new_x, new_y
end

function initialize_player_dependend_information(players, planets, stars)
	local home_systems = {}
	local nr_players = 0
    for i, player in pairs(players) do
		local home_system = stars[planets[player.home_planet].star]
		nr_players = nr_players + 1
		table.insert(home_systems, home_system)
	end
	return nr_players, home_systems
end

function is_outside_of_valid_galaxy_box(x, y, max_x, max_y)
	if x < 0 + MINIMUM_DISTANCE_FROM_GALAXY_EDGE_IN_COORDINATES 
	or x > max_x - MINIMUM_DISTANCE_FROM_GALAXY_EDGE_IN_COORDINATES
	or y < 0 + MINIMUM_DISTANCE_FROM_GALAXY_EDGE_IN_COORDINATES 
	or y > max_y - MINIMUM_DISTANCE_FROM_GALAXY_EDGE_IN_COORDINATES then 
		return true
	else
		return false
	end
end

function swap_star_coordinates(star_a, star_b)
	local tmp_x, tmp_y = star_a.x, star_a.y
	star_a.x, star_a.y = star_b.x, star_b.y
	star_b.x, star_b.y = tmp_x, tmp_y
	return star_a, star_b
end

function is_a_home_system(star, home_systems)
	local is_home_system = false
	for i, home_system in pairs(home_systems) do
		if star == home_system then
			return true
		end
	end
	return false
end

function get_next_valid_star_id(stars, next_star_id, home_systems)
	if next_star_id >= #stars - 1 then
		return -1
	end
	while next_star_id < #stars - 1 do
		next_star_id = next_star_id + 1
		local star = stars[next_star_id]
		local orbits = star.orbits
		--somehow #star.orbits does not work work
		local nr_orbits = 0
		for _ in pairs(orbits) do
			nr_orbits = nr_orbits + 1
		end
		if is_a_home_system(star, home_systems) or (star.color == BLACK_HOLE_COLOUR and DELETE_ALL_BLACKHOLES) or (nr_orbits == 0 and DELETE_ALL_EMPTY_SYSTEMS) then
		else
			break;
		end
	end
	return next_star_id
end

function go()
	set_stars({}) -- early check that we're in universal context (inspired by hwmirror.lua)

	local stars = get_stars()
		
	local players = get_players()
	local planets = get_planets()
	
	local nr_players, home_systems = initialize_player_dependend_information(players, planets, stars)
	
	local game = get_game()
	local max_x = game.map_max_x
	local max_y = game.map_max_y	
	local center_x = max_x/2
	local center_y = max_y/2
	local nr_stars = #stars + 1

	--assumption: max_y <= max_x
	local outer_cluster_radius_threshold = max_y * CLUSTER_SIZE_LIMIT_IN_VERTICAL_PERCENTAGE / 200
	local arm_degrees_per_player = 360 / nr_players
	
	-- was multiplied by 1 mio
	local cosinus = {[0]=1000000, [1]=999848, [2]=999391, [3]=998630, [4]=997564, [5]=996195, [6]=994522, [7]=992546, [8]=990268, [9]=987688, [10]=984808, [11]=981627, [12]=978148, [13]=974370, [14]=970296, [15]=965926, [16]=961262, [17]=956305, [18]=951057, [19]=945519, [20]=939693, [21]=933580, [22]=927184, [23]=920505, [24]=913545, [25]=906308, [26]=898794, [27]=891007, [28]=882948, [29]=874620, [30]=866025, [31]=857167, [32]=848048, [33]=838671, [34]=829038, [35]=819152, [36]=809017, [37]=798636, [38]=788011, [39]=777146, [40]=766044, [41]=754710, [42]=743145, [43]=731354, [44]=719340, [45]=707107, [46]=694658, [47]=681998, [48]=669131, [49]=656059, [50]=642788, [51]=629320, [52]=615661, [53]=601815, [54]=587785, [55]=573576, [56]=559193, [57]=544639, [58]=529919, [59]=515038, [60]=500000, [61]=484810, [62]=469472, [63]=453990, [64]=438371, [65]=422618, [66]=406737, [67]=390731, [68]=374607, [69]=358368, [70]=342020, [71]=325568, [72]=309017, [73]=292372, [74]=275637, [75]=258819, [76]=241922, [77]=224951, [78]=207912, [79]=190809, [80]=173648, [81]=156434, [82]=139173, [83]=121869, [84]=104528, [85]=87156, [86]=69756, [87]=52336, [88]=34899, [89]=17452, [90]=0, [91]=-17452, [92]=-34899, [93]=-52336, [94]=-69756, [95]=-87156, [96]=-104528, [97]=-121869, [98]=-139173, [99]=-156434, [100]=-173648, [101]=-190809, [102]=-207912, [103]=-224951, [104]=-241922, [105]=-258819, [106]=-275637, [107]=-292372, [108]=-309017, [109]=-325568, [110]=-342020, [111]=-358368, [112]=-374607, [113]=-390731, [114]=-406737, [115]=-422618, [116]=-438371, [117]=-453990, [118]=-469472, [119]=-484810, [120]=-500000, [121]=-515038, [122]=-529919, [123]=-544639, [124]=-559193, [125]=-573576, [126]=-587785, [127]=-601815, [128]=-615661, [129]=-629320, [130]=-642788, [131]=-656059, [132]=-669131, [133]=-681998, [134]=-694658, [135]=-707107, [136]=-719340, [137]=-731354, [138]=-743145, [139]=-754710, [140]=-766044, [141]=-777146, [142]=-788011, [143]=-798636, [144]=-809017, [145]=-819152, [146]=-829038, [147]=-838671, [148]=-848048, [149]=-857167, [150]=-866025, [151]=-874620, [152]=-882948, [153]=-891007, [154]=-898794, [155]=-906308, [156]=-913545, [157]=-920505, [158]=-927184, [159]=-933580, [160]=-939693, [161]=-945519, [162]=-951057, [163]=-956305, [164]=-961262, [165]=-965926, [166]=-970296, [167]=-974370, [168]=-978148, [169]=-981627, [170]=-984808, [171]=-987688, [172]=-990268, [173]=-992546, [174]=-994522, [175]=-996195, [176]=-997564, [177]=-998630, [178]=-999391, [179]=-999848, [180]=-1000000, [181]=-999848, [182]=-999391, [183]=-998630, [184]=-997564, [185]=-996195, [186]=-994522, [187]=-992546, [188]=-990268, [189]=-987688, [190]=-984808, [191]=-981627, [192]=-978148, [193]=-974370, [194]=-970296, [195]=-965926, [196]=-961262, [197]=-956305, [198]=-951057, [199]=-945519, [200]=-939693, [201]=-933580, [202]=-927184, [203]=-920505, [204]=-913545, [205]=-906308, [206]=-898794, [207]=-891007, [208]=-882948, [209]=-874620, [210]=-866025, [211]=-857167, [212]=-848048, [213]=-838671, [214]=-829038, [215]=-819152, [216]=-809017, [217]=-798636, [218]=-788011, [219]=-777146, [220]=-766044, [221]=-754710, [222]=-743145, [223]=-731354, [224]=-719340, [225]=-707107, [226]=-694658, [227]=-681998, [228]=-669131, [229]=-656059, [230]=-642788, [231]=-629320, [232]=-615661, [233]=-601815, [234]=-587785, [235]=-573576, [236]=-559193, [237]=-544639, [238]=-529919, [239]=-515038, [240]=-500000, [241]=-484810, [242]=-469472, [243]=-453990, [244]=-438371, [245]=-422618, [246]=-406737, [247]=-390731, [248]=-374607, [249]=-358368, [250]=-342020, [251]=-325568, [252]=-309017, [253]=-292372, [254]=-275637, [255]=-258819, [256]=-241922, [257]=-224951, [258]=-207912, [259]=-190809, [260]=-173648, [261]=-156434, [262]=-139173, [263]=-121869, [264]=-104528, [265]=-87156, [266]=-69756, [267]=-52336, [268]=-34899, [269]=-17452, [270]=-0, [271]=17452, [272]=34899, [273]=52336, [274]=69756, [275]=87156, [276]=104528, [277]=121869, [278]=139173, [279]=156434, [280]=173648, [281]=190809, [282]=207912, [283]=224951, [284]=241922, [285]=258819, [286]=275637, [287]=292372, [288]=309017, [289]=325568, [290]=342020, [291]=358368, [292]=374607, [293]=390731, [294]=406737, [295]=422618, [296]=438371, [297]=453990, [298]=469472, [299]=484810, [300]=500000, [301]=515038, [302]=529919, [303]=544639, [304]=559193, [305]=573576, [306]=587785, [307]=601815, [308]=615661, [309]=629320, [310]=642788, [311]=656059, [312]=669131, [313]=681998, [314]=694658, [315]=707107, [316]=719340, [317]=731354, [318]=743145, [319]=754710, [320]=766044, [321]=777146, [322]=788011, [323]=798636, [324]=809017, [325]=819152, [326]=829038, [327]=838671, [328]=848048, [329]=857167, [330]=866025, [331]=874620, [332]=882948, [333]=891007, [334]=898794, [335]=906308, [336]=913545, [337]=920505, [338]=927184, [339]=933580, [340]=939693, [341]=945519, [342]=951057, [343]=956305, [344]=961262, [345]=965926, [346]=970296, [347]=974370, [348]=978148, [349]=981627, [350]=984808, [351]=987688, [352]=990268, [353]=992546, [354]=994522, [355]=996195, [356]=997564, [357]=998630, [358]=999391, [359]=999848}
	local sinus = {[0]=0, [1]=17452, [2]=34899, [3]=52336, [4]=69756, [5]=87156, [6]=104528, [7]=121869, [8]=139173, [9]=156434, [10]=173648, [11]=190809, [12]=207912, [13]=224951, [14]=241922, [15]=258819, [16]=275637, [17]=292372, [18]=309017, [19]=325568, [20]=342020, [21]=358368, [22]=374607, [23]=390731, [24]=406737, [25]=422618, [26]=438371, [27]=453990, [28]=469472, [29]=484810, [30]=500000, [31]=515038, [32]=529919, [33]=544639, [34]=559193, [35]=573576, [36]=587785, [37]=601815, [38]=615661, [39]=629320, [40]=642788, [41]=656059, [42]=669131, [43]=681998, [44]=694658, [45]=707107, [46]=719340, [47]=731354, [48]=743145, [49]=754710, [50]=766044, [51]=777146, [52]=788011, [53]=798636, [54]=809017, [55]=819152, [56]=829038, [57]=838671, [58]=848048, [59]=857167, [60]=866025, [61]=874620, [62]=882948, [63]=891007, [64]=898794, [65]=906308, [66]=913545, [67]=920505, [68]=927184, [69]=933580, [70]=939693, [71]=945519, [72]=951057, [73]=956305, [74]=961262, [75]=965926, [76]=970296, [77]=974370, [78]=978148, [79]=981627, [80]=984808, [81]=987688, [82]=990268, [83]=992546, [84]=994522, [85]=996195, [86]=997564, [87]=998630, [88]=999391, [89]=999848, [90]=1000000, [91]=999848, [92]=999391, [93]=998630, [94]=997564, [95]=996195, [96]=994522, [97]=992546, [98]=990268, [99]=987688, [100]=984808, [101]=981627, [102]=978148, [103]=974370, [104]=970296, [105]=965926, [106]=961262, [107]=956305, [108]=951057, [109]=945519, [110]=939693, [111]=933580, [112]=927184, [113]=920505, [114]=913545, [115]=906308, [116]=898794, [117]=891007, [118]=882948, [119]=874620, [120]=866025, [121]=857167, [122]=848048, [123]=838671, [124]=829038, [125]=819152, [126]=809017, [127]=798636, [128]=788011, [129]=777146, [130]=766044, [131]=754710, [132]=743145, [133]=731354, [134]=719340, [135]=707107, [136]=694658, [137]=681998, [138]=669131, [139]=656059, [140]=642788, [141]=629320, [142]=615661, [143]=601815, [144]=587785, [145]=573576, [146]=559193, [147]=544639, [148]=529919, [149]=515038, [150]=500000, [151]=484810, [152]=469472, [153]=453990, [154]=438371, [155]=422618, [156]=406737, [157]=390731, [158]=374607, [159]=358368, [160]=342020, [161]=325568, [162]=309017, [163]=292372, [164]=275637, [165]=258819, [166]=241922, [167]=224951, [168]=207912, [169]=190809, [170]=173648, [171]=156434, [172]=139173, [173]=121869, [174]=104528, [175]=87156, [176]=69756, [177]=52336, [178]=34899, [179]=17452, [180]=0, [181]=-17452, [182]=-34899, [183]=-52336, [184]=-69756, [185]=-87156, [186]=-104528, [187]=-121869, [188]=-139173, [189]=-156434, [190]=-173648, [191]=-190809, [192]=-207912, [193]=-224951, [194]=-241922, [195]=-258819, [196]=-275637, [197]=-292372, [198]=-309017, [199]=-325568, [200]=-342020, [201]=-358368, [202]=-374607, [203]=-390731, [204]=-406737, [205]=-422618, [206]=-438371, [207]=-453990, [208]=-469472, [209]=-484810, [210]=-500000, [211]=-515038, [212]=-529919, [213]=-544639, [214]=-559193, [215]=-573576, [216]=-587785, [217]=-601815, [218]=-615661, [219]=-629320, [220]=-642788, [221]=-656059, [222]=-669131, [223]=-681998, [224]=-694658, [225]=-707107, [226]=-719340, [227]=-731354, [228]=-743145, [229]=-754710, [230]=-766044, [231]=-777146, [232]=-788011, [233]=-798636, [234]=-809017, [235]=-819152, [236]=-829038, [237]=-838671, [238]=-848048, [239]=-857167, [240]=-866025, [241]=-874620, [242]=-882948, [243]=-891007, [244]=-898794, [245]=-906308, [246]=-913545, [247]=-920505, [248]=-927184, [249]=-933580, [250]=-939693, [251]=-945519, [252]=-951057, [253]=-956305, [254]=-961262, [255]=-965926, [256]=-970296, [257]=-974370, [258]=-978148, [259]=-981627, [260]=-984808, [261]=-987688, [262]=-990268, [263]=-992546, [264]=-994522, [265]=-996195, [266]=-997564, [267]=-998630, [268]=-999391, [269]=-999848, [270]=-1000000, [271]=-999848, [272]=-999391, [273]=-998630, [274]=-997564, [275]=-996195, [276]=-994522, [277]=-992546, [278]=-990268, [279]=-987688, [280]=-984808, [281]=-981627, [282]=-978148, [283]=-974370, [284]=-970296, [285]=-965926, [286]=-961262, [287]=-956305, [288]=-951057, [289]=-945519, [290]=-939693, [291]=-933580, [292]=-927184, [293]=-920505, [294]=-913545, [295]=-906308, [296]=-898794, [297]=-891007, [298]=-882948, [299]=-874620, [300]=-866025, [301]=-857167, [302]=-848048, [303]=-838671, [304]=-829038, [305]=-819152, [306]=-809017, [307]=-798636, [308]=-788011, [309]=-777146, [310]=-766044, [311]=-754710, [312]=-743145, [313]=-731354, [314]=-719340, [315]=-707107, [316]=-694658, [317]=-681998, [318]=-669131, [319]=-656059, [320]=-642788, [321]=-629320, [322]=-615661, [323]=-601815, [324]=-587785, [325]=-573576, [326]=-559193, [327]=-544639, [328]=-529919, [329]=-515038, [330]=-500000, [331]=-484810, [332]=-469472, [333]=-453990, [334]=-438371, [335]=-422618, [336]=-406737, [337]=-390731, [338]=-374607, [339]=-358368, [340]=-342020, [341]=-325568, [342]=-309017, [343]=-292372, [344]=-275637, [345]=-258819, [346]=-241922, [347]=-224951, [348]=-207912, [349]=-190809, [350]=-173648, [351]=-156434, [352]=-139173, [353]=-121869, [354]=-104528, [355]=-87156, [356]=-69756, [357]=-52336, [358]=-34899, [359]=-17452}

	-- updated during generation, will be swapped with home worlds
	local outermost_arm_systems = {}
	local fallback_outermost_arm_system_next_player_id = 1

	-- Apparently id 0 is always orion if Orion exists
	next_star_id = get_next_valid_star_id(stars, -1, home_systems)
	local center_star = stars[next_star_id]
	center_star.keep = true
	center_star.x = center_x
	center_star.y = center_y
	local current_cluster_radius = 0
	while current_cluster_radius <= outer_cluster_radius_threshold do
	
		current_cluster_radius = current_cluster_radius + CLUSTER_DISTANCE_BETWEEN_RINGS_IN_COORDINATES
		local ring_circumference = 2 * current_cluster_radius * 314159 / 100000
	
		local nr_stars_in_ring = ring_circumference / CLUSTER_DISTANCE_BETWEEN_RING_STARS_IN_COORDINATES
		local angle_per_ring_star = 360 / nr_stars_in_ring

		for i = 1, nr_stars_in_ring do
			next_star_id = get_next_valid_star_id(stars, next_star_id, home_systems)
			local star = stars[next_star_id]
			star.keep = true
			local angle = i * angle_per_ring_star + CLUSTER_ROTATION_ANGLE
			star.x, star.y = offset_to_coordinates(center_x, center_y, angle, current_cluster_radius, sinus, cosinus)

			outermost_arm_systems[fallback_outermost_arm_system_next_player_id] = star
			fallback_outermost_arm_system_next_player_id = fallback_outermost_arm_system_next_player_id + 1
			if fallback_outermost_arm_system_next_player_id > nr_players then
				fallback_outermost_arm_system_next_player_id = 1
			end
		end

	end
	
	-- INNERMOST ARM SYSTEMS
	local previous_arm_system_position_x = {}
	local previous_arm_system_position_y = {}
	for i = 1, nr_players do
		local angle = i * arm_degrees_per_player + ARM_ROTATION_ANGLE
		previous_arm_system_position_x[i], previous_arm_system_position_y[i] = offset_to_coordinates(center_x, center_y, angle, current_cluster_radius + ARM_DISTANCE_TO_CLUSTER_IN_COORDINATES, sinus, cosinus)		

		local full_circumfence = 2 * current_cluster_radius * 314159 / 100000
		local circumfence_covered = (ARM_INNER_CIRCUMFENCE_COVERED_FACTOR_PER_PLAYER * full_circumfence) / (100 * nr_players)
		
		--replacing with method still has some issue: next_star_id, stars = place_stars_in_arm_iteration_part(full_circumfence, circumfence_covered, stars, next_star_id, home_systems, previous_arm_system_position_x, previous_arm_system_position_y, center_x, center_y, angle, current_cluster_radius, sinus, cosinus)

		local circumfence_based_angle_offset_between_stars = (ARM_DISTANCE_BETWEEN_PARTIAL_RING_STARS_IN_COORDINATES * 360) / full_circumfence
		
		local nr_stars_to_place = 1 + 2 * circumfence_covered / ARM_DISTANCE_BETWEEN_PARTIAL_RING_STARS_IN_COORDINATES

		local uneven_nr_of_stars = nr_stars_to_place % 2 == 1

		if uneven_nr_of_stars then
			next_star_id = get_next_valid_star_id(stars, next_star_id, home_systems)
			if next_star_id == -1 then
				msgbox("SPIRAL:\nERROR: Failed at innermost galaxy am generation: \nNot enough stars to produce a proper spiral. \nIncrease galaxy size, reduce number of players or tweak mod parameters.")
			end
			local star = stars[next_star_id]
			star.keep = true
			star.x, star.y = previous_arm_system_position_x[i], previous_arm_system_position_y[i]

			outermost_arm_systems[i] = star
		end

		for j = 1, nr_stars_to_place / 2 do
			local angle_deviation = circumfence_based_angle_offset_between_stars * j
			if not uneven_nr_of_stars then
				angle_deviation = angle_deviation - circumfence_based_angle_offset_between_stars / 2
			end

			local left_x, left_y = offset_to_coordinates(center_x, center_y, angle - angle_deviation / 2, current_cluster_radius + ARM_DISTANCE_TO_CLUSTER_IN_COORDINATES, sinus, cosinus)
			local right_x, right_y = offset_to_coordinates(center_x, center_y, angle + angle_deviation / 2, current_cluster_radius + ARM_DISTANCE_TO_CLUSTER_IN_COORDINATES, sinus, cosinus)

			next_star_id = get_next_valid_star_id(stars, next_star_id, home_systems)
			if next_star_id == -1 then
				msgbox("SPIRAL:\nERROR: Failed at innermost galaxy am generation: \nNot enough stars to produce a proper spiral. \nIncrease galaxy size, reduce number of players or tweak mod parameters.")
			end
			local left_star = stars[next_star_id]
			left_star.keep = true
			left_star.x, left_star.y = left_x, left_y

			outermost_arm_systems[i] = left_star
			
			next_star_id = get_next_valid_star_id(stars, next_star_id, home_systems)
			if next_star_id ~= -1 then
				local right_star = stars[next_star_id]
				right_star.keep = true
				right_star.x, right_star.y = right_x, right_y

				outermost_arm_systems[i] = right_star
			end
		end
	end
	
	-- OUTER ARM SYSTEMS
	local arm_iteration = 0
	local remaining_line_distance_factor = ARM_OUTER_LINE_COVERED_STARTING_FACTOR_PER_PLAYER
	local remaining_factor_reduction_root = ARM_OUTER_LINE_COVERED_INITIAL_ROOTED_FACTOR_LOSS_PER_PLAYER
	local placement_is_valid = true
	current_cluster_radius = current_cluster_radius - CLUSTER_DISTANCE_BETWEEN_RINGS_IN_COORDINATES
	local previous_angle_for_player = {}
	for i = 1, nr_players do
		previous_angle_for_player[i] = i * arm_degrees_per_player + INITIAL_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION + ARM_ROTATION_ANGLE
	end
	while placement_is_valid == true do
		arm_iteration = arm_iteration + 1
		
		local arm_system_coordinates_x = {}
		local arm_system_coordinates_y = {}
		
		placement_is_valid = true
		
		local full_circumfence = 2 * current_cluster_radius * 314159 / 100000
		if full_circumfence <= 0 then
			--msgbox("todo: warning at full circumfence (galaxy to small)")
			full_circumfence = 1
		end
		local distance_covered = (remaining_line_distance_factor * full_circumfence) / (100 * nr_players)
		if distance_covered < 0 then
			distance_covered = 0
		end
		remaining_line_distance_factor = remaining_line_distance_factor - remaining_factor_reduction_root * remaining_factor_reduction_root
		remaining_factor_reduction_root = remaining_factor_reduction_root - 1
		if remaining_factor_reduction_root < 1 then
			remaining_factor_reduction_root = 1
		end
		
		local circumfence_based_angle_offset_between_stars = (ARM_DISTANCE_BETWEEN_PARTIAL_RING_STARS_IN_COORDINATES * 360) / full_circumfence
		
		local nr_stars_to_place = 1 + 2 * distance_covered / ARM_DISTANCE_BETWEEN_PARTIAL_RING_STARS_IN_COORDINATES

		local uneven_nr_of_stars = nr_stars_to_place % 2 == 1
		
		for i = 1, nr_players do
			local angle = previous_angle_for_player[i] + arm_iteration_to_angle_change(arm_iteration, INITIAL_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE, FINAL_SPIRAL_ARM_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_FINAL_SPIRAL_ARM_ANGLE_CHANGE)
			if arm_iteration == 1 then
				--compensate switch from curvature to straight line
				arm_system_coordinates_x[i], arm_system_coordinates_y[i] = offset_to_coordinates(previous_arm_system_position_x[i], previous_arm_system_position_y[i], angle, ARM_DISTANCE_INNER_TO_FIRST_OUTER_IN_COORDINATES, sinus, cosinus)
			else
				arm_system_coordinates_x[i], arm_system_coordinates_y[i] = offset_to_coordinates(previous_arm_system_position_x[i], previous_arm_system_position_y[i], angle, ARM_STAR_DISTANCE_IN_COORDINATES, sinus, cosinus)
			end
			if uneven_nr_of_stars then
				if is_outside_of_valid_galaxy_box(arm_system_coordinates_x[i], arm_system_coordinates_y[i], max_x, max_y) then
					if placement_is_valid then
						msgbox("SPIRAL:\nNote: stopped growing arms because edge of galaxy was reached.")
					end
					placement_is_valid = false
				end
			end
			for j = 1, nr_stars_to_place / 2 do
				local left_x, left_y   = offset_to_coordinates(arm_system_coordinates_x[i], arm_system_coordinates_y[i], angle-90, (ARM_STAR_DISTANCE_IN_COORDINATES * j) / 2, sinus, cosinus)
				local right_x, right_y = offset_to_coordinates(arm_system_coordinates_x[i], arm_system_coordinates_y[i], angle+90, (ARM_STAR_DISTANCE_IN_COORDINATES * j) / 2, sinus, cosinus)

				if is_outside_of_valid_galaxy_box(left_x, left_y, max_x, max_y) or is_outside_of_valid_galaxy_box(right_x, right_y, max_x, max_y) then
					if placement_is_valid then
						msgbox("SPIRAL:\nNote: stopped growing arms because edge of galaxy was reached.")
					end
					placement_is_valid = false
				end
			end
		end

		if arm_iteration > NR_MAX_OUTER_ARM_ITERATIONS then
			if placement_is_valid then
				msgbox("SPIRAL:\nNote: stopped growing arms because maxmimum defined numbe of iterations was reached.")
			end
			placement_is_valid = false
		end

		local remembered_next_star_id = next_star_id
		for i = 1, nr_players * nr_stars_to_place do
			next_star_id = get_next_valid_star_id(stars, next_star_id, home_systems)
			if next_star_id == -1 then
				placement_is_valid = false
				break
			end
		end
		next_star_id = remembered_next_star_id
		
		if placement_is_valid then
			for i = 1, nr_players do
				local angle = previous_angle_for_player[i] + arm_iteration_to_angle_change(arm_iteration, INITIAL_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_INTERMEDIATE_SPIRAL_ARM_ANGLE_CHANGE, FINAL_SPIRAL_ARM_CHANGE_PER_ITERATION, NR_ITERATIONS_TO_REACH_FINAL_SPIRAL_ARM_ANGLE_CHANGE)

				if uneven_nr_of_stars then
					next_star_id = get_next_valid_star_id(stars, next_star_id, home_systems)
					local star = stars[next_star_id]
					star.keep = true
					star.x, star.y = arm_system_coordinates_x[i], arm_system_coordinates_y[i]
					outermost_arm_systems[i] = star
				end

				for j = 1, nr_stars_to_place / 2 do
					local offset_to_middle_of_arm_iteration = (ARM_STAR_DISTANCE_IN_COORDINATES * j)
					if not uneven_nr_of_stars then
						offset_to_middle_of_arm_iteration = offset_to_middle_of_arm_iteration - ARM_STAR_DISTANCE_IN_COORDINATES / 2
					end
					local left_x, left_y   = offset_to_coordinates(arm_system_coordinates_x[i], arm_system_coordinates_y[i], angle-90, offset_to_middle_of_arm_iteration, sinus, cosinus)
					local right_x, right_y = offset_to_coordinates(arm_system_coordinates_x[i], arm_system_coordinates_y[i], angle+90, offset_to_middle_of_arm_iteration, sinus, cosinus)

					next_star_id = get_next_valid_star_id(stars, next_star_id, home_systems)

					local left_star = stars[next_star_id]
					left_star.keep = true
					left_star.x, left_star.y = left_x, left_y
					outermost_arm_systems[i] = left_star
					
					next_star_id = get_next_valid_star_id(stars, next_star_id, home_systems)

					if next_star_id ~= -1 then
						local right_star = stars[next_star_id]
						right_star.keep = true
						right_star.x, right_star.y = right_x, right_y
					end
				end
				previous_arm_system_position_x[i], previous_arm_system_position_y[i] = arm_system_coordinates_x[i], arm_system_coordinates_y[i]
				previous_angle_for_player[i] = angle
			end
		else
			break
		end
	end
	
	if REMOVE_ALL_WORMHOLES then
		for i, star in pairs(stars) do
			star.wormhole = -1
		end
	end

	for i = 1, nr_players do
		local home_system = home_systems[i]
		local outermost_arm_system = outermost_arm_systems[i]
		swap_star_coordinates(outermost_arm_systems[i], home_systems[i])
		home_system.keep = true
		outermost_arm_system.keep = nil
	end
	
	local stars_to_delete = {}
	for i, star in pairs(stars) do
		if star.keep == nil or star.keep ~= true then
			table.insert(stars_to_delete, i)
		else
			star.keep = nil
		end
	end
	set_stars(stars)
		
	msgbox("SPIRAL:\nNote: deleting " .. #stars_to_delete .. " systems that are not used.") --todo count those!
	del_stars(stars_to_delete)
end

go()